"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AWS = require("aws-sdk");
const _debug = require("debug");
const _ = require("lodash");
const debug = _debug('ecs-updater');
const TASKDEFINITION_SUFFIX = '_taskdefinition.json';
const TAG_SUFFIX = '_tag';
const REVISION_SUFFIX = '_revision';
let _ecs;
function getECS(region) {
    if (!_ecs) {
        _ecs = new AWS.ECS({ region });
        // _ecs = promisifyMethods(awsecs, [
        //   'describeServices',
        //   'registerTaskDefinition',
        //   'updateService',
        //   'describeTaskDefinition',
        // ]);
    }
    return _ecs;
}
let _s3;
function getS3(region) {
    if (!_s3) {
        _s3 = new AWS.S3({ region });
        // _s3 = promisifyMethods(awss3, [
        //   'getObject',
        //   'putObject',
        // ]);
    }
    return _s3;
}
function getRegisteredTaskDefinition(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const ecs = getECS(config.REGION);
        const result = yield ecs.describeServices({
            cluster: config.CLUSTER,
            services: [config.SERVICE],
        }).promise();
        const tdResponse = yield ecs.describeTaskDefinition({
            taskDefinition: result.services[0].taskDefinition,
        }).promise();
        return tdResponse.taskDefinition;
    });
}
exports.getRegisteredTaskDefinition = getRegisteredTaskDefinition;
function getTaskDefinition(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const taskDefinition = yield getS3TaskDefinition(config);
        if (taskDefinition) {
            console.log(`[INFO] Using taskDefinition from s3`);
            return taskDefinition;
        }
        return getRegisteredTaskDefinition(config);
    });
}
exports.getTaskDefinition = getTaskDefinition;
function getS3TaskDefinition(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const taskDefinition = yield getS3Object(config, TASKDEFINITION_SUFFIX, config.TASKDEFINITION_KEY);
        if (taskDefinition && typeof taskDefinition !== 'string') {
            return taskDefinition;
        }
        return undefined;
    });
}
exports.getS3TaskDefinition = getS3TaskDefinition;
function getS3ImageTag(config) {
    return getS3Object(config, TAG_SUFFIX);
}
exports.getS3ImageTag = getS3ImageTag;
function registerTaskDefinition(config, _taskDefinition) {
    let taskDefinition = _taskDefinition;
    if (isRegistered(_taskDefinition)) {
        taskDefinition = registeredToVanilla(_taskDefinition);
    }
    const ecs = getECS(config.REGION);
    return ecs.registerTaskDefinition(taskDefinition).promise()
        .then(result => result.taskDefinition);
}
exports.registerTaskDefinition = registerTaskDefinition;
function registeredToVanilla(definition) {
    return {
        family: definition.family,
        containerDefinitions: definition.containerDefinitions,
        volumes: definition.volumes,
    };
}
function isRegistered(definition) {
    return definition.taskDefinitionArn !== undefined;
}
exports.isRegistered = isRegistered;
function restartService(config, taskDefinition) {
    const ecs = getECS(config.REGION);
    return ecs.updateService({
        cluster: config.CLUSTER,
        service: config.SERVICE,
        taskDefinition: taskDefinition.taskDefinitionArn,
    }).promise();
}
exports.restartService = restartService;
function deploy(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const config = overrideValues(opts, {
            REGION: '',
            CLUSTER: '',
            SERVICE: '',
            CONTAINER: '',
            IMAGE_TAG: '',
        });
        const missingValues = falseyKeys(config);
        if (missingValues.length) {
            throw new Error('Error: All configuration values are required.  ' +
                'Missing values: ' + missingValues.join(', '));
        }
        if (opts.BUCKET) {
            config.BUCKET = opts.BUCKET;
        }
        if (opts.KEY) {
            config.KEY = opts.KEY;
        }
        if (opts.TASKDEFINITION_KEY) {
            config.TASKDEFINITION_KEY = opts.TASKDEFINITION_KEY;
        }
        const currentTaskDefinition = yield getTaskDefinition(config);
        const currentContainer = getContainer(config.CONTAINER, currentTaskDefinition);
        const nextContainer = Object.assign({}, currentContainer, { image: config.IMAGE_TAG ? updateTag(currentContainer.image, config.IMAGE_TAG) : currentContainer.image });
        const nextTaskDefinition = nextTask(currentTaskDefinition, nextContainer);
        const registeredTaskDefinition = yield registerTaskDefinition(config, nextTaskDefinition);
        yield restartService(config, registeredTaskDefinition);
        return {
            taskDefinition: registeredTaskDefinition,
            container: nextContainer,
        };
    });
}
exports.deploy = deploy;
function terraformRestart(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const config = overrideValues(opts, {
            REGION: '',
            CLUSTER: '',
            SERVICE: '',
            CONTAINER: '',
            BUCKET: '',
            KEY: '',
        });
        const missingValues = falseyKeys(config);
        if (missingValues.length) {
            throw new Error('Error: All configuration values are required.  ' +
                'Missing values: ' + missingValues.join(', '));
        }
        if (opts.TASKDEFINITION_KEY) {
            config.TASKDEFINITION_KEY = opts.TASKDEFINITION_KEY;
        }
        const template = yield getS3TaskDefinition(config);
        if (!template) {
            throw new Error("Couldn't find taskDefinition");
        }
        const imageTag = yield getS3ImageTag(config);
        if (!imageTag) {
            throw new Error("Couldn't find image tag from S3");
        }
        const currentContainer = getContainer(config.CONTAINER, template);
        const nextContainer = Object.assign({}, currentContainer, { image: updateTag(currentContainer.image, imageTag) });
        const nextTaskDefinition = nextTask(template, nextContainer);
        const registeredTaskDefinition = yield registerTaskDefinition(config, nextTaskDefinition);
        yield restartService(config, registeredTaskDefinition);
        return {
            taskDefinition: registeredTaskDefinition,
            container: nextContainer,
        };
    });
}
exports.terraformRestart = terraformRestart;
function restart(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const config = overrideValues(opts, {
            REGION: '',
            CLUSTER: '',
            SERVICE: '',
        });
        const missingValues = falseyKeys(config);
        if (missingValues.length) {
            throw new Error('Error: All configuration values are required.  ' +
                'Missing values: ' + missingValues.join(', '));
        }
        const template = yield getRegisteredTaskDefinition(config);
        const registeredTaskDefinition = yield registerTaskDefinition(config, template);
        yield restartService(config, registeredTaskDefinition);
        return {
            current: registeredTaskDefinition,
            previous: template,
        };
    });
}
exports.restart = restart;
function getContainer(containerName, taskDefinition) {
    const containers = taskDefinition.containerDefinitions.filter(container => container.name === containerName);
    if (containers.length === 0) {
        throw new Error(`The taskDefinition didn't have a container named ${containerName}`);
    }
    return containers[0];
}
function nextTask(taskDefinition, nextContainer) {
    return Object.assign({}, taskDefinition, { containerDefinitions: taskDefinition.containerDefinitions.map((container) => {
            if (container.name === nextContainer.name) {
                return nextContainer;
            }
            return container;
        }) });
}
function updateTag(image, tag) {
    const parts = /^(https?:\/\/)?([\w-_.\/]+)(:\w+)?$/.exec(image);
    if (!parts) {
        throw new Error(`Invalid container image: ${image}`);
    }
    const newImage = (parts[1] || '') + parts[2] + ':' + tag;
    debug('Replacing image %s with %s', image, newImage);
    return newImage;
}
exports.updateTag = updateTag;
function syncRevision(config, taskDefinition) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!config.BUCKET || !config.KEY) {
            throw new Error(`Can't syncRevision since BUCKET or KEY is not set`);
        }
        const S3 = getS3(config.REGION);
        const revision = taskDefinition.revision;
        const revisionKey = config.KEY + REVISION_SUFFIX;
        yield S3.putObject({
            Bucket: config.BUCKET,
            Key: revisionKey,
            ContentType: 'text/plain',
            Body: String(revision),
        });
        console.log(`s3://${config.BUCKET}/${revisionKey} => ${revision}`);
    });
}
exports.syncRevision = syncRevision;
function syncImageTag(config, container) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!config.BUCKET || !config.KEY) {
            throw new Error(`Can't syncImageTag since BUCKET or KEY is not set`);
        }
        const S3 = getS3(config.REGION);
        const tag = getTag(container.image);
        const tagKey = config.KEY + '_tag';
        yield S3.putObject({
            Bucket: config.BUCKET,
            Key: tagKey,
            ContentType: 'text/plain',
            Body: tag,
        });
        console.log(`s3://${config.BUCKET}/${tagKey} => ${tag}`);
    });
}
exports.syncImageTag = syncImageTag;
function getImage(imageWithTag) {
    const parts = imageWithTag.split(':');
    parts.pop();
    return parts.join(':');
}
function getTag(imageWithTag) {
    const parts = imageWithTag.split(':');
    return parts[parts.length - 1];
}
function overrideValues(overrides, defaults) {
    return _.assign({}, defaults, _.pick(overrides, _.keys(defaults)));
}
exports.overrideValues = overrideValues;
function falseyKeys(obj) {
    return Object.keys(obj).filter(k => obj[k] === '');
}
function upsert(array, keys, item) {
    const id = _.pick(item, keys);
    return _.without(array, _.find(array, id)).concat(item);
}
function getS3Object(config, suffix, key) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!config.BUCKET || !config.KEY) {
            return undefined;
        }
        const S3 = getS3(config.REGION);
        const Key = key ? key : config.KEY + suffix;
        try {
            const response = yield S3.getObject({
                Bucket: config.BUCKET,
                Key,
            }).promise();
            let body = response.Body;
            if (Buffer.isBuffer(body)) {
                body = body.toString();
            }
            if (_.isString(body)) {
                if (response.ContentType && response.ContentType.indexOf('json') !== -1) {
                    return JSON.parse(body);
                }
                return body;
            }
        }
        catch (err) {
            console.log(err);
            console.log(`[INFO] Couldn\'t find s3://${config.BUCKET}/${Key}`);
        }
        return undefined;
    });
}
//# sourceMappingURL=ecs-deploy.js.map