#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const AWS = require("aws-sdk");
const chalk = require("chalk");
const child_process_1 = require("child_process");
const program = require("commander");
const _debug = require("debug");
const util_1 = require("util");
const config_1 = require("./config");
const deployer = require("./ecs-deploy");
const debug = _debug('ecs-updater');
require('pkginfo')(module, 'version');
program
    .version(module.exports.version)
    .usage('[options]')
    .description(`Build, tag and upload a Docker image and then restart an ECS service.
  Optionally specify a sub-command.`)
    .option('-s, --sub-command <which>', 'login|build|restart-service|restart-terraform|taskDefinition', /^(login|build|restart\-service|restart\-terraform|taskDefinition)$/)
    .option('--no-login', 'Skip login')
    .parse(process.argv);
const opts = program.opts();
switch (opts.subCommand) {
    case 'login':
        ecrLogin(config_1.default)
            .then(dockerLogin)
            .then(console.log, fail());
        break;
    case 'build':
        ecrLogin(config_1.default)
            .then(r => build(config_1.default, tryPrependRepo(config_1.default.IMAGE, r.endpoint)))
            .then(console.log, fail());
        break;
    case 'restart-service':
        restartService(config_1.default)
            .then(logObject, fail());
        break;
    case 'restart-terraform':
        terraformRestart(config_1.default)
            .then(logObject, fail());
        break;
    case 'taskDefinition':
        deployer.getTaskDefinition(config_1.default)
            .then(current => deployer.registerTaskDefinition(config_1.default, current))
            .then(logObject, fail());
        break;
    default:
        if (!opts.subCommand) {
            start(config_1.default, opts.login)
                .catch(fail());
        }
        else {
            console.error('Invalid syntax');
        }
        break;
}
function fail(msg) {
    if (msg) {
        console.log(msg);
    }
    return (err) => {
        console.log(chalk.bold.red('\nFailed with error:\n'));
        console.log(err);
        process.exit(1);
    };
}
function logObject(obj, depth = 4) {
    const _inspect = util_1.inspect;
    console.log(_inspect(obj, { depth, colors: true, maxArrayLength: 10 }));
}
function restartService(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const { current, previous } = yield deployer.restart(config);
        yield deployer.syncRevision(config, current);
        return {
            current,
            previous,
        };
    });
}
exports.restartService = restartService;
function terraformRestart(config) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log(chalk.bold.green('\nRestarting the service to use the latest taskDefinition in S3\n'));
        debug(config);
        const { container, taskDefinition } = yield deployer.terraformRestart(config);
        console.log('Updated service %s to revision %s', config.SERVICE, taskDefinition.revision);
        console.log(chalk.bold.green('\nSyncing revision and image tag to S3\n'));
        yield deployer.syncRevision(config, taskDefinition);
        yield deployer.syncImageTag(config, container);
        return {
            container,
            taskDefinition,
        };
    });
}
exports.terraformRestart = terraformRestart;
function runDocker(...args) {
    return __awaiter(this, void 0, void 0, function* () {
        const stdio = 'inherit';
        let silent = false;
        if (args[0] === 'silent') {
            args.shift();
            silent = true;
        }
        return new Promise((resolve, reject) => {
            const docker = child_process_1.spawn('docker', args, { stdio });
            if (!silent) {
                console.log(['docker'].concat(args).join(' '));
            }
            docker.on('close', (code) => {
                if (code !== 0) {
                    console.log(`docker process exited with code ${code}`);
                    reject(code);
                    return;
                }
                resolve(true);
            });
            docker.on('error', (err) => {
                console.log(`docker process exited with code ${err}`);
                reject(err);
            });
        });
    });
}
function ecrLogin(config) {
    return __awaiter(this, void 0, void 0, function* () {
        const ecr = new AWS.ECR({ region: config.REGION });
        const response = yield ecr.getAuthorizationToken().promise();
        if (!response.authorizationData) {
            throw new Error('Invalid response');
        }
        const o = response.authorizationData[0];
        const login = Buffer.from(o.authorizationToken, 'base64').toString('utf8');
        const parts = login.split(':');
        return {
            user: parts[0],
            password: parts[1],
            endpoint: o.proxyEndpoint,
        };
    });
}
exports.ecrLogin = ecrLogin;
function dockerLogin({ user, password, endpoint }) {
    return runDocker('silent', 'login', '-u', user, '-p', password, '-e', 'none', endpoint);
}
function build(config, image) {
    return runDocker('build', '-f', config.DOCKERFILE, '-t', `${image || config.IMAGE}:${config.IMAGE_TAG}`, '.');
}
function push(config, image) {
    return runDocker('push', `${image || config.IMAGE}:${config.IMAGE_TAG}`);
}
function login(config) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log(chalk.bold.green('Logging in to ECR\n'));
        // get ECR login
        const credentials = yield ecrLogin(config);
        // login with Docker
        yield dockerLogin(credentials);
        return tryPrependRepo(config.IMAGE, credentials.endpoint);
    });
}
function getRevision(taskDefinitionArn) {
    const parts = taskDefinitionArn.split(':');
    return parseInt(parts[parts.length - 1], 10);
}
function isECR(image) {
    return image.indexOf('/') === -1; // we assume Docker Hub otherwise
}
function tryPrependRepo(image, endpoint) {
    let newImage = image;
    if (isECR(image)) {
        const parts = endpoint.split('//');
        if (parts.length === 2) {
            newImage = parts[1] + '/' + image;
        }
    }
    return newImage;
}
function start(config, loginFlag = true) {
    return __awaiter(this, void 0, void 0, function* () {
        let image = config.IMAGE;
        if (loginFlag && isECR(config.IMAGE)) {
            image = yield login(config);
        }
        console.log(chalk.bold.green('\nBuilding the Docker image\n'));
        // build
        yield build(config, image);
        console.log(chalk.bold.green('\nPushing to repository\n'));
        yield push(config, image);
        // deploy service
        console.log(chalk.bold.green('\nRestarting ECS service %s\n'), config.SERVICE);
        const configCopy = Object.assign({}, config, { IMAGE: image });
        const { container, taskDefinition } = yield deployer.deploy(configCopy);
        console.log('Updated service %s to use image %s', config.SERVICE, container.image);
        if (config.BUCKET && config.KEY) {
            console.log(chalk.bold.green('\nUpdating revision and image tag to S3\n'));
            yield deployer.syncRevision(configCopy, taskDefinition);
            yield deployer.syncImageTag(configCopy, container);
        }
        console.log(chalk.bold.green('\nDONE'));
    });
}
exports.start = start;
//# sourceMappingURL=index.js.map